#include <math.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <stdlib.h>

#include "application.h"
#include "calcs.h"
#include "messages.h"
#include "qplotwidget.h"
#include "ddplot.h"


// Calculation of the distance between two atoms given by vectors atom1 and atom2
double vDistance( double *v1, double *v2, int nitems=3 )
{
  double dist;
  int d;

  dist = 0;
  for (d=0; d<nitems; d++)
    dist += pow(v1[d]-v2[d],2);
  dist = sqrt(dist);

  return dist;
}


double vDotProduct( double v1[], double v2[], int nitems=3 )
{
  double res;
  int n;

  res = 0;
  for (n=0; n<nitems; n++)
    res += v1[n]*v2[n];

  return res;
}


void swap( double &val1, double &val2 )
{
  double keep;

  keep = val1;
  val1 = val2;
  val2 = keep; 
}


//---------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------


QPlotWidget::QPlotWidget(QWidget *parent, const char *name, WFlags f) 
  : QWidget( parent, name, f )
{
  int d;

  xBorder = yBorder = 30;
  xOffset = yOffset = 0;
  arrAngle = 60.0;
  ZFact = ZOOM_FACT;
  xPan = yPan = 0;
  isSelected = true;
  DisloCenter = SHOW_DCENTER;

  for (d=0; d<3; d++) 
    ProjVector[d] = PROJ_VECTOR[d];

  DScaleFact = DSCALE_FACT;
  shortestArrow = ARROW_SHORTEST;     // shortest arrow to be plotted
  thicknessArrow = ARROW_THICKNESS;
  plotType = PLOT_TYPE;
  DispComponent = DISP_COMPONENT;
  AtomPos = ATOMPOS;
  AtomDispScale = ATOMDISP_SCALE;
  PrintMagDisp = PRINT_MAG_DISP;
  AtomNumbers = ATOM_NUMBERS;

  setGeometry( 0, 23, parent->width(), parent->height()-23 );
}


void QPlotWidget::paintEvent( QPaintEvent *)
{
  QPainter *p = new QPainter(this);
  drawIt(p);
  p->end();
}


void QPlotWidget::drawIt( QPainter *p )
{
  QString txt;
  double x,y;
  int n,layer,prev_layer,circDia;

  QPen *pen = new QPen();
  QBrush *brush = new QBrush();
  QBrush *brush_unrld = new QBrush();

  if (AtomPos==COMPOSITE) {
    brush_unrld->setColor( lightGray );
    brush_unrld->setStyle( SolidPattern );
  }

  // plotting of atomic arrangement in the initial configuration
  prev_layer = -1;
  for (n=0; n<NInit; n++) {
    if (plotType==PLOT_ATOM_LAYERS)
      layer = zLayer[n]; 
    else
      layer = atomType[n];

    if (layer!=prev_layer) {
      pen->setColor( zColorLayer[layer][0] );
      pen->setWidth( zLineThickLayer[layer] );
      p->setPen( *pen );

      brush->setColor( zColorLayer[layer][1] );
      brush->setStyle( SolidPattern );
      p->setBrush( *brush );   

      circDia = zDiamLayer[layer];
      prev_layer = layer;
    }

    // plot the atomic positions

    switch(AtomPos) {
      case UNRELAXED:
        x = ZFact*xyzInit[n][0]-xPan-(ZFact-1)*xCen;
	y = ZFact*xyzInit[n][1]-yPan-(ZFact-1)*yCen;
        DrawCircle(p,n,x,y,circDia);
        break;

      case RELAXED:
        x = ZFact*(xyzInit[n][0]+AtomDispScale*aDisp[n][0])-xPan-(ZFact-1)*xCen;
        y = ZFact*(xyzInit[n][1]+AtomDispScale*aDisp[n][1])-yPan-(ZFact-1)*yCen;
        DrawCircle(p,n,x,y,circDia);
	break;

      case COMPOSITE:
	p->setBrush( *brush_unrld );
	x = ZFact*xyzInit[n][0]-xPan-(ZFact-1)*xCen;
	y = ZFact*xyzInit[n][1]-yPan-(ZFact-1)*yCen;
        DrawCircle(p,n,x,y,circDia);
  
        p->setBrush( *brush );
	x = ZFact*(xyzInit[n][0]+AtomDispScale*aDisp[n][0])-xPan-(ZFact-1)*xCen;
	y = ZFact*(xyzInit[n][1]+AtomDispScale*aDisp[n][1])-yPan-(ZFact-1)*yCen; 
        DrawCircle(p,-1,x,y,circDia);
    }

    // atom numbers
    if (AtomNumbers) {
      txt.sprintf("%d",n);
      DrawText(p,x,y,txt);
    }
  }

  // plotting of arrows
  QPen apen( black, thicknessArrow );
  p->setPen( apen ); 
  p->setBrush( Qt::black );
   
  switch(DispComponent) {
    case EDGE:
      plotEdgeComponent(p);
      break;
    case SCREW:
      plotScrewComponent(p);
      break;
    case MIXED:
      plotEdgeComponent(p);
  }

  // position of dislocation center
  if (DisloCenter) {
    x = ZFact*xCore-xPan-(ZFact-1)*xCen;
    y = ZFact*yCore-yPan-(ZFact-1)*yCen;
    DrawCircle(p,-1,x,y,circDia);
  }
}


void QPlotWidget::printIt( QPrinter *printer )
{
  QPainter *prn = new QPainter(printer);
  QString cmd;
  fpos_t *pos;
  char txt[80];
  int err;

  drawIt(prn);
  prn->end();

  // fix the bounding box                                                                  
  if (printer->outputToFile()) {
#ifdef Q_WS_X11  // X11 specific                                             
    /*
    cmd = "gs  -sDEVICE=bbox -dNOPAUSE -dQUIET -dBATCH " + printer->outputFileName();

    system(cmd.latin1());
    fflush(stdout);
    fsetpos(stdout,pos);

    fscanf(stdout,"%s\n",txt);
    cout << txt << endl;
    */
#elif defined(Q_WS_MACX)   // MacOS X specific                                             

#endif
  }
}


//---------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------


// Plotting of the edge component of the displacements
void QPlotWidget::plotEdgeComponent(QPainter *p)
{
  double arr_between[MAX_ATOMS][MAX_NEIGHBORS];
  double rdisp[3];
  double xy_from[2],xy_to[2];
  double cen,lproj,pvsize;
  int i,d,in,ineigh,icen,idx;
  bool exist;


  for (i=0; i<MAX_ATOMS; i++)
    arr_between[i][0] = END_OF_LIST;

  // plotting of arrows
  for (icen=0; icen<NInit; icen++) {
    idx = in = 0;    

    while (aNeighbors[icen][idx]!=END_OF_LIST) {
      ineigh = aNeighbors[icen][idx];

      // does the arrow between icen<-ineigh already exist ?
      i = 0;
      exist = false;
      while (arr_between[ineigh][i]!=END_OF_LIST) {
        if (arr_between[ineigh][i]==icen) {
          exist = true;
          break;
        }
        i++;
      }

      // yes, the arrow exists - get out
      if (exist) {
        idx++;
        continue;
      }
   
      // arrow does not exist - plot it
      arr_between[icen][in] = ineigh;                    // the arrow is between icen and ineigh
      in++;

      //----- from here it holds only for edges and mixed dislocations -----      
      switch(DispComponent) {
        case EDGE:
          for (d=0; d<2; d++)
            rdisp[d] = -aDisp[ineigh][d]+aDisp[icen][d];      // relative displacement of the two atoms
          break;

        case MIXED:
	  lproj = 0;
          for (d=0; d<3; d++)
            lproj += (-aDisp[ineigh][d]+aDisp[icen][d])*ProjVector[d]; 
 
          pvsize = sqrt( pow(ProjVector[0],2)+pow(ProjVector[1],2)+pow(ProjVector[2],2) );
          lproj /= pvsize;    // lproj is now the intercept on the (unit length) projection vector

          for (d=0; d<2; d++)
            rdisp[d] = (-aDisp[ineigh][d]+aDisp[icen][d])*lproj;  
      }
// ^^^^^^^^^^^ this should be aDisp[ineigh][d]-aDisp[icen][d], but the arrows are then in the other
// ^^^^^^^^^^^ direction than in the plotting program under Windows

      // calculation of the length of an arrow between icen and ineigh
      for (d=0; d<2; d++) {
	if (AtomPos==UNRELAXED)
          cen = (xyzInit[icen][d]+xyzInit[ineigh][d])/2.0;         // center of the arrow
        else
          cen = (xyzRel[icen][d]+AtomDispScale*aDisp[icen][d]+
            xyzRel[ineigh][d]+AtomDispScale*aDisp[ineigh][d])/2.0;         // center of the arrow

        xy_from[d] = cen - rdisp[d]/2.0*DScaleFact;
        xy_to[d] = cen + rdisp[d]/2.0*DScaleFact;
      }
      //--------------------------------------------------------------------

      // plotting of the arrows
      DrawArrow( p, ZFact*xy_from[0]-xPan-(ZFact-1)*xCen, ZFact*xy_from[1]-yPan-(ZFact-1)*yCen, 
        ZFact*xy_to[0]-xPan-(ZFact-1)*xCen, ZFact*xy_to[1]-yPan-(ZFact-1)*yCen);
      idx++;
    }

    arr_between[icen][in] = END_OF_LIST;               // terminator
  }
}


// Plotting of the screw component of the displacements
void QPlotWidget::plotScrewComponent(QPainter *p)
{
  QString txt;
  double arr_between[MAX_ATOMS][MAX_NEIGHBORS];
  double rdisp[3],xyz0[3],xyz1[3],rvect[3];
  double xyz_from[3],xyz_to[3];
  double rscreen[3];
  double rproj,ratio,cen;
  int i,d,in,ineigh,icen,idx;
  bool exist,plotted;


  for (i=0; i<MAX_ATOMS; i++)
    arr_between[i][0] = END_OF_LIST;

  // plotting of arrows
  for (icen=0; icen<NInit; icen++) {
    idx = in = 0;    

    while (aNeighbors[icen][idx]!=END_OF_LIST) {
      ineigh = aNeighbors[icen][idx];

      // does the arrow between icen<-ineigh already exist ?
      i = 0;
      exist = false;
      while (arr_between[ineigh][i]!=END_OF_LIST) {
        if (arr_between[ineigh][i]==icen) {
          exist = true;
          break;
        }
        i++;
      }

      // yes, the arrow exists - get out
      if (exist) {
        idx++;
        continue;
      }
   
      // arrow does not exist - plot it
      arr_between[icen][in] = ineigh;                    // the arrow is between icen and ineigh
      in++;
      
      for (d=0; d<3; d++)
        rdisp[d] = aDisp[ineigh][d]-aDisp[icen][d];      // relative displacement of the two atoms

      //----- from here it holds only for screws -----
      rproj = rdisp[2];                                   // z-displacement
      ratio = rproj/lenBurgers;
      
      if (fabs(ratio)>1.0)                                // treatment if ratio>1
        ratio = ratio-int(ratio);      

      if (fabs(ratio)>0.5)
        ratio -= fabs(ratio)/ratio;

      // calculation of the length of an arrow between icen and ineigh
      if (AtomPos==UNRELAXED) {
        for (d=0; d<3; d++) {
          xyz0[d] = xyzInit[icen][d];
          xyz1[d] = xyzInit[ineigh][d]; 
        }
      } else {
        for (d=0; d<3; d++) {
          xyz0[d] = xyzRel[icen][d] + AtomDispScale*aDisp[icen][d];
          xyz1[d] = xyzRel[ineigh][d] + AtomDispScale*aDisp[ineigh][d]; 
        }
      }
      for (d=0; d<3; d++) {
        rvect[d] = xyz1[d]-xyz0[d];
	rscreen[d] = ratio*2*DScaleFact*rvect[d];            // length of the arrow (from -1 to 1)
      }

      // adjustment of arrows - they must lie between two atoms
      for (d=0; d<3; d++) {
        cen = (xyz0[d]+xyz1[d])/2.0;
        xyz_from[d] = cen - rscreen[d]/2.0;
        xyz_to[d] = cen + rscreen[d]/2.0;
      }
      //----------------------------------------------

      // plotting of the arrows
      plotted = DrawArrow( p, ZFact*xyz_from[0]-xPan-(ZFact-1)*xCen, ZFact*xyz_from[1]-yPan-(ZFact-1)*yCen, 
        ZFact*xyz_to[0]-xPan-(ZFact-1)*xCen, ZFact*xyz_to[1]-yPan-(ZFact-1)*yCen);

      // annotations
      if (PrintMagDisp && plotted) {
        txt.sprintf("a/%0.1lf",1.0/(fabs(ratio)/2.0));
        DrawText( p, ZFact*(xyz_from[0]+xyz_to[0])/2.0, ZFact*(xyz_from[1]+xyz_to[1])/2.0, txt );
      }
 
      idx++;
    }

    arr_between[icen][in] = END_OF_LIST;            // terminator
  }
}


//---------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------


bool QPlotWidget::DrawArrow( QPainter *p, double x0, double y0, double x1, double y1 )
{
  QPointArray pts;
  const double h = 0.2*tan(arrAngle*M_PI/360.0);
  double arrow[4][2] = { {0.0,0.0}, {1.0,0.0}, {0.8,h}, {0.8,-h} };
  int i,xy[4][2];
  double ang,xrot,yrot,size;

  size = sqrt(pow(x1-x0,2)+pow(y1-y0,2));
  if (factor*size<shortestArrow) 
    return false;

  // scaling and rotation
  ang = atan2(y1-y0,x1-x0);
  for (i=0; i<4; i++) {
    // scaling
    arrow[i][0] *= size;
    arrow[i][1] *= size;

    // rotation
    xrot = arrow[i][0]*cos(ang)-arrow[i][1]*sin(ang);
    yrot = arrow[i][0]*sin(ang)+arrow[i][1]*cos(ang);
    arrow[i][0] = xrot;
    arrow[i][1] = yrot;
  }

  for (i=0; i<4; i++) {
    xy[i][0] = xBorder + xOffset + (int) ceil(factor*(x0+arrow[i][0]-xMin));
    xy[i][1] = height() - (yBorder + yOffset + (int) ceil(factor*(y0+arrow[i][1]-yMin)));
  }

  p->drawLine(xy[0][0],xy[0][1],xy[1][0],xy[1][1]);
  pts.setPoints(3,xy[1][0],xy[1][1],xy[2][0],xy[2][1],xy[3][0],xy[3][1]);
  p->drawPolygon( pts );
  return true;
}


void QPlotWidget::DrawCircle( QPainter *p, int atom, double x, double y, int dia)
{
  int xpos,ypos;
  int rad = int(dia/2.0);

  // (0,0) is at the bottom left corner of the screen
  xpos = xBorder + xOffset + (int) floor(factor*(x-xMin)) - rad; 
  ypos = height() - (yBorder + yOffset + (int) floor(factor*(y-yMin))) - rad;
  
  p->drawEllipse( xpos, ypos, dia, dia );

  // if the real atomic number is given, keep its coordinates on the screen
  if (atom>=0) {
    atomScreen[atom][0] = xpos + rad;
    atomScreen[atom][1] = ypos + rad;
  }
}


void QPlotWidget::DrawLine( QPainter *p, double x0, double y0, double x1, double y1, int addx, int addy)
{
  int xfrom,xto,yfrom,yto;

  xfrom = xBorder + xOffset + (int) floor(factor*(x0-xMin)) + addx;
  xto   = xBorder + xOffset + (int) floor(factor*(x1-xMin)) + addx;
  yfrom = height() - (yBorder + yOffset + (int) floor(factor*(y0-yMin)) + addy);
  yto   = height() - (yBorder + yOffset + (int) floor(factor*(y1-yMin)) + addy);

  p->drawLine( xfrom, yfrom, xto, yto );
}


void QPlotWidget::DrawText( QPainter *p, double x, double y, QString txt )
{
  int xf,yf;

  xf = xBorder + xOffset + (int) floor(factor*(x-xMin));
  yf = height() - (yBorder + yOffset + (int) floor(factor*(y-yMin)));

  p->drawText( xf, yf, txt );
}


void QPlotWidget::Pan( double xsteps, double ysteps )
{
  xPan += PAN_RATIO*xsteps*xSize;   
  yPan += PAN_RATIO*ysteps*ySize;   
  repaint();
}


void QPlotWidget::SetGeometry()
{
  double xfact,yfact;

  xfact = (width()-2*xBorder)/xSize;
  yfact = (height()-2*yBorder)/ySize;
  factor = (xfact<yfact) ? xfact : yfact;

  xOffset = (int) floor( (width()-2*xBorder-factor*xSize)/2.0 );
  yOffset = (int) floor( (height()-2*yBorder-factor*ySize)/2.0 );
}


void QPlotWidget::Zoom( double zfact )
{
  if (ZFact*zfact<=20 && ZFact*zfact>=0.1) {
    ZFact *= zfact;
    repaint();
  }
}


//---------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------


// Finds the atom in the topmost layer that is nearest to the position [x,y] on the screen
int QPlotWidget::FindNearestAtom( int x, int y )
{
  double dist,mindist;
  int n,atom;

  atom = -1;
  for (n=0; n<NInit; n++) {
    dist = sqrt( pow(atomScreen[n][0]-x,2) + pow(atomScreen[n][1]-y,2) );

    if (dist<mindist || atom<0) {
      mindist = dist;
      atom = n;
    }
  }

  return atom;
}


// Seeking the nearest neighbors
bool QPlotWidget::FindNeighbors()
{
  double dist;
  int icen,ineigh,idx;
  bool ok;

  ok = true;

  for (icen=0; icen<NInit; icen++) {
    idx = 0;
    for (ineigh=0; ineigh<NInit; ineigh++) {
      if (icen==ineigh)
        continue;

      dist = sqrt( pow(xyzInit[icen][0]-xyzInit[ineigh][0],2) + 
        pow(xyzInit[icen][1]-xyzInit[ineigh][1],2) );
      
      if (dist<=MAX_NEIGHBOR_DIST) {
        if (idx>=MAX_NEIGHBORS) {
          msgCritical( crit_MAX_NEIGHBORS_Exceeded );
          icen = ineigh = NInit;
          break;
        }
          
        aNeighbors[icen][idx] = ineigh;
        idx++;
      }
    }        
    aNeighbors[icen][idx] = END_OF_LIST;
  }

  return ok;
}


// Reading the plot file
bool QPlotWidget::LoadFile( QString fname )
{
  const char *fn = fname.ascii(); 
  FILE *f;
  double unused;
  int n;

  FName = fname;
  f = fopen(fn,"r+");
  if (f==NULL) {
    printf("Error:  Input data file %s cannot be found.\n",fn);
    return false;
  }

  // relaxed block - reading of coordinates

  fscanf(f,"%d\n",&NRel);
  if (NRel>MAX_ATOMS) {
    msgCritical( crit_MAX_ATOMS_Exceeded );
    fclose(f);
    return false;
  }

  for (n=0; n<NRel; n++)
    fscanf(f, "%lf\n", &xyzRel[n][2]);
  for (n=0; n<NRel; n++)
    fscanf(f, "%lf %lf %d\n", &xyzRel[n][0], &xyzRel[n][1], &atomType[n]);

  // initial block - reading of coordinates

  fscanf(f,"%d\n",&NInit);
  if (NInit>MAX_ATOMS) {
    msgCritical( crit_MAX_ATOMS_Exceeded );
    fclose(f);
    return false;
  }

  for (n=0; n<NInit; n++)
    fscanf(f, "%lf\n", &xyzInit[n][2]);
  for (n=0; n<NInit; n++)
    fscanf(f, "%lf %lf\n", &xyzInit[n][0], &xyzInit[n][1]);

  fscanf(f, "%lf \n %lf \n %lf \n %lf \n", &unused, &unused, &unused, &lenBurgers);
  fscanf(f, "CORE { %lf;%lf", &xCore, &yCore);

  fclose(f);

  return true;
}


// Calculation of the atomic displacements
void QPlotWidget::PrepareBlock()
{ 
  double zcoord[MAX_ATOMS][2];
  double zlast;
  int m,n,d,layer,fgidx,bgidx,thidx,diaidx;

  // calculation of the boundaries of the block of atoms

  xMin = yMin = zMin = 32767;
  xMax = yMax = zMax = -32767;
  for (n=0; n<NInit; n++) {
    if (xyzInit[n][0]<xMin)  xMin = xyzInit[n][0];
    if (xyzInit[n][1]<yMin)  yMin = xyzInit[n][1];
    if (xyzInit[n][2]<zMin)  zMin = xyzInit[n][2];
    if (xyzInit[n][0]>xMax)  xMax = xyzInit[n][0];
    if (xyzInit[n][1]>yMax)  yMax = xyzInit[n][1];    
    if (xyzInit[n][2]>zMax)  zMax = xyzInit[n][2]; 
  }
  xSize = xMax-xMin;
  ySize = yMax-yMin;  
  xCen = (int) trunc((xMin+xMax)/2.0);
  yCen = (int) trunc((yMin+yMax)/2.0);
  zCen = (int) trunc((zMin+zMax)/2.0);

  // looking for the layers

  for (n=0; n<NInit; n++) {
    zcoord[n][0] = n;
    zcoord[n][1] = xyzInit[n][2];
  }

  for (m=0; m<NInit; m++) {
    for (n=0; n<NInit; n++) {
      if (zcoord[m][1]<zcoord[n][1]) {
        swap(zcoord[m][0],zcoord[n][0]);
        swap(zcoord[m][1],zcoord[n][1]);
      }  
    }
  }

  // layers and their appearance

  layer = -1;
  for (n=0; n<NInit; n++) {
    if (n==0 || zcoord[n][1]>zlast) {
      layer++;     
      zlast = zcoord[n][1];

      zCoordLayer[layer] = zlast;

      // calculation of the color, thickness and diameter
      fgidx = CalcIndex(layer,NUM_FGCOLORS);
      bgidx = CalcIndex(layer,NUM_BGCOLORS);
      thidx = CalcIndex(layer,NUM_THICKNESS);
      diaidx = CalcIndex(layer,NUM_DIAM);

      zColorLayer[layer][0] = ATOM_FGCOLOR[fgidx];
      zColorLayer[layer][1] = ATOM_BGCOLOR[bgidx];
      zLineThickLayer[layer] = ATOM_THICKNESS[thidx];
      zDiamLayer[layer] = ATOM_DIAM[diaidx];
    }
    zLayer[(int) zcoord[n][0]] = layer;
  } 
  NumZLayers = layer+1;

  // calculation of the displacement vectors

  for (n=0; n<NRel; n++) {
    for (d=0; d<3; d++)
      aDisp[n][d] = xyzRel[n][d]-xyzInit[n][d];
  }

  // looking for neighbors
  FindNeighbors();
}


